@@ -10,10 +10,12 @@
 #include <private/qnoncontiguousbytedevice_p.h>
 
 #include <QtNetwork/qabstractsocket.h>
+
 #include <QtCore/qloggingcategory.h>
 #include <QtCore/qendian.h>
 #include <QtCore/qdebug.h>
 #include <QtCore/qlist.h>
+#include <QtCore/qnumeric.h>
 #include <QtCore/qurl.h>
 
 #include <qhttp2configuration.h>
@@ -86,30 +88,6 @@ HPack::HttpHeader build_headers(const QHttpNetworkRequest &request, quint32 maxH
     return header;
 }
 
-std::vector<uchar> assemble_hpack_block(const std::vector<Http2::Frame> &frames)
-{
-    std::vector<uchar> hpackBlock;
-
-    quint32 total = 0;
-    for (const auto &frame : frames)
-        total += frame.hpackBlockSize();
-
-    if (!total)
-        return hpackBlock;
-
-    hpackBlock.resize(total);
-    auto dst = hpackBlock.begin();
-    for (const auto &frame : frames) {
-        if (const auto hpackBlockSize = frame.hpackBlockSize()) {
-            const uchar *src = frame.hpackBlockBegin();
-            std::copy(src, src + hpackBlockSize, dst);
-            dst += hpackBlockSize;
-        }
-    }
-
-    return hpackBlock;
-}
-
 QUrl urlkey_from_request(const QHttpNetworkRequest &request)
 {
     QUrl url;
@@ -309,8 +287,7 @@ bool QHttp2ProtocolHandler::sendRequest()
     auto &requests = m_channel->h2RequestsToSend;
     for (auto it = requests.begin(), endIt = requests.end(); it != endIt;) {
         const auto &pair = *it;
-        const QString scheme(pair.first.url().scheme());
-        if (scheme == "preconnect-http"_L1 || scheme == "preconnect-https"_L1) {
+        if (pair.first.isPreConnect()) {
             m_connection->preConnectFinished();
             emit pair.second->finished();
             it = requests.erase(it);
@@ -348,11 +325,13 @@ bool QHttp2ProtocolHandler::sendRequest()
         initReplyFromPushPromise(message, key);
     }
 
-    const auto streamsToUse = std::min<quint32>(maxConcurrentStreams > quint32(activeStreams.size())
-                                                ? maxConcurrentStreams - quint32(activeStreams.size()) : 0,
-                                                requests.size());
+    const auto isClientSide = [](const auto &pair) -> bool { return (pair.first & 1) == 1; };
+    const auto activeClientSideStreams = std::count_if(
+            activeStreams.constKeyValueBegin(), activeStreams.constKeyValueEnd(), isClientSide);
+    const qint64 streamsToUse = qBound(0, qint64(maxConcurrentStreams) - activeClientSideStreams,
+                                       requests.size());
     auto it = requests.begin();
-    for (quint32 i = 0; i < streamsToUse; ++i) {
+    for (qint64 i = 0; i < streamsToUse; ++i) {
         const qint32 newStreamID = createNewStream(*it);
         if (!newStreamID) {
             // TODO: actually we have to open a new connection.
@@ -959,7 +938,7 @@ void QHttp2ProtocolHandler::handleContinuedHEADERS()
         // has yet to see the reset.
     }
 
-    std::vector<uchar> hpackBlock(assemble_hpack_block(continuedFrames));
+    std::vector<uchar> hpackBlock(Http2::assemble_hpack_block(continuedFrames));
     const bool hasHeaderFields = !hpackBlock.empty();
     if (hasHeaderFields) {
         HPack::BitIStream inputStream{&hpackBlock[0], &hpackBlock[0] + hpackBlock.size()};
@@ -1098,7 +1077,7 @@ void QHttp2ProtocolHandler::updateStream(Stream &stream, const HPack::HttpHeader
     int statusCode = 0;
     for (const auto &pair : headers) {
         const auto &name = pair.name;
-        auto value = pair.value;
+        const auto value = QByteArrayView(pair.value);
 
         // TODO: part of this code copies what SPDY protocol handler does when
         // processing headers. Binary nature of HTTP/2 and SPDY saves us a lot
@@ -1118,67 +1097,8 @@ void QHttp2ProtocolHandler::updateStream(Stream &stream, const HPack::HttpHeader
             if (ok)
                 httpReply->setContentLength(length);
         } else {
-            QByteArray binder(", ");
-            if (name == "set-cookie")
-                binder = "\n";
-            httpReply->appendHeaderField(name, value.replace('\0', binder));
-        }
-    }
-
-    const auto handleAuth = [&, this](QByteArrayView authField, bool isProxy) -> bool {
-        Q_ASSERT(httpReply);
-        const QByteArrayView auth = authField.trimmed();
-        if (auth.startsWith("Negotiate") || auth.startsWith("NTLM")) {
-            // @todo: We're supposed to fall back to http/1.1:
-            // https://docs.microsoft.com/en-us/iis/get-started/whats-new-in-iis-10/http2-on-iis#when-is-http2-not-supported
-            // "Windows authentication (NTLM/Kerberos/Negotiate) is not supported with HTTP/2.
-            // In this case IIS will fall back to HTTP/1.1."
-            // Though it might be OK to ignore this. The server shouldn't let us connect with
-            // HTTP/2 if it doesn't support us using it.
-        } else if (!auth.isEmpty()) {
-            // Somewhat mimics parts of QHttpNetworkConnectionChannel::handleStatus
-            bool resend = false;
-            const bool authenticateHandled = m_connection->d_func()->handleAuthenticateChallenge(
-                    m_socket, httpReply, isProxy, resend);
-            if (authenticateHandled && resend) {
-                httpReply->d_func()->eraseData();
-                // Add the request back in queue, we'll retry later now that
-                // we've gotten some username/password set on it:
-                httpRequest.d->needResendWithCredentials = true;
-                m_channel->h2RequestsToSend.insert(httpRequest.priority(), stream.httpPair);
-                httpReply->d_func()->clearHeaders();
-                // If we have data we were uploading we need to reset it:
-                if (stream.data()) {
-                    stream.data()->reset();
-                    httpReplyPrivate->totallyUploadedData = 0;
-                }
-                return true;
-            } // else: Authentication failed or was cancelled
-        }
-        return false;
-    };
-
-    if (httpReply) {
-        // See Note further down. These statuses would in HTTP/1.1 be handled
-        // by QHttpNetworkConnectionChannel::handleStatus. But because h2 has
-        // multiple streams/requests in a single channel this structure does not
-        // map properly to that function.
-        if (httpReply->statusCode() == 401) {
-            const auto wwwAuth = httpReply->headerField("www-authenticate");
-            if (handleAuth(wwwAuth, false)) {
-                sendRST_STREAM(stream.streamID, CANCEL);
-                markAsReset(stream.streamID);
-                // The stream is finalized and deleted after returning
-                return;
-            } // else: errors handled later
-        } else if (httpReply->statusCode() == 407) {
-            const auto proxyAuth = httpReply->headerField("proxy-authenticate");
-            if (handleAuth(proxyAuth, true)) {
-                sendRST_STREAM(stream.streamID, CANCEL);
-                markAsReset(stream.streamID);
-                // The stream is finalized and deleted after returning
-                return;
-            } // else: errors handled later
+            const auto binder = name == "set-cookie" ? QByteArrayView("\n") : QByteArrayView(", ");
+            httpReply->appendHeaderField(name, QByteArray(pair.value).replace('\0', binder));
         }
     }
 
@@ -1256,6 +1176,91 @@ void QHttp2ProtocolHandler::updateStream(Stream &stream, const Frame &frame,
     }
 }
 
+// After calling this function, either the request will be re-sent or
+// the reply will be finishedWithError! Do not emit finished() or similar on the
+// reply after this!
+void QHttp2ProtocolHandler::handleAuthorization(Stream &stream)
+{
+    auto *httpReply = stream.reply();
+    auto *httpReplyPrivate = httpReply->d_func();
+    auto &httpRequest = stream.request();
+
+    Q_ASSERT(httpReply && (httpReply->statusCode() == 401 || httpReply->statusCode() == 407));
+
+    const auto handleAuth = [&, this](QByteArrayView authField, bool isProxy) -> bool {
+        Q_ASSERT(httpReply);
+        const QByteArrayView auth = authField.trimmed();
+        if (auth.startsWith("Negotiate") || auth.startsWith("NTLM")) {
+            // @todo: We're supposed to fall back to http/1.1:
+            // https://docs.microsoft.com/en-us/iis/get-started/whats-new-in-iis-10/http2-on-iis#when-is-http2-not-supported
+            // "Windows authentication (NTLM/Kerberos/Negotiate) is not supported with HTTP/2.
+            // In this case IIS will fall back to HTTP/1.1."
+            // Though it might be OK to ignore this. The server shouldn't let us connect with
+            // HTTP/2 if it doesn't support us using it.
+            return false;
+        }
+        // Somewhat mimics parts of QHttpNetworkConnectionChannel::handleStatus
+        bool resend = false;
+        const bool authenticateHandled = m_connection->d_func()->handleAuthenticateChallenge(
+                m_socket, httpReply, isProxy, resend);
+        if (authenticateHandled) {
+            if (resend) {
+                httpReply->d_func()->eraseData();
+                // Add the request back in queue, we'll retry later now that
+                // we've gotten some username/password set on it:
+                httpRequest.d->needResendWithCredentials = true;
+                m_channel->h2RequestsToSend.insert(httpRequest.priority(), stream.httpPair);
+                httpReply->d_func()->clearHeaders();
+                // If we have data we were uploading we need to reset it:
+                if (stream.data()) {
+                    stream.data()->reset();
+                    httpReplyPrivate->totallyUploadedData = 0;
+                }
+                // We automatically try to send new requests when the stream is
+                // closed, so we don't need to call sendRequest ourselves.
+                return true;
+            } // else: we're just not resending the request.
+            // @note In the http/1.x case we (at time of writing) call close()
+            // for the connectionChannel (which is a bit weird, we could surely
+            // reuse the open socket outside "connection:close"?), but in http2
+            // we only have one channel, so we won't close anything.
+        } else {
+            // No authentication header or authentication isn't supported, but
+            // we got a 401/407 so we cannot succeed. We need to emit signals
+            // for headers and data, and then finishWithError.
+            emit httpReply->headerChanged();
+            emit httpReply->readyRead();
+            QNetworkReply::NetworkError error = httpReply->statusCode() == 401
+                    ? QNetworkReply::AuthenticationRequiredError
+                    : QNetworkReply::ProxyAuthenticationRequiredError;
+            finishStreamWithError(stream, QNetworkReply::AuthenticationRequiredError,
+                                  m_connection->d_func()->errorDetail(error, m_socket));
+        }
+        return false;
+    };
+
+    // These statuses would in HTTP/1.1 be handled by
+    // QHttpNetworkConnectionChannel::handleStatus. But because h2 has
+    // multiple streams/requests in a single channel this structure does not
+    // map properly to that function.
+    bool authOk = true;
+    switch (httpReply->statusCode()) {
+    case 401:
+        authOk = handleAuth(httpReply->headerField("www-authenticate"), false);
+        break;
+    case 407:
+        authOk = handleAuth(httpReply->headerField("proxy-authenticate"), true);
+        break;
+    default:
+        Q_UNREACHABLE();
+    }
+    if (authOk) {
+        markAsReset(stream.streamID);
+        deleteActiveStream(stream.streamID);
+    } // else: errors handled inside handleAuth
+}
+
+// Called when we have received a frame with the END_STREAM flag set
 void QHttp2ProtocolHandler::finishStream(Stream &stream, Qt::ConnectionType connectionType)
 {
     Q_ASSERT(stream.state == Stream::remoteReserved || stream.reply());
@@ -1263,6 +1268,15 @@ void QHttp2ProtocolHandler::finishStream(Stream &stream, Qt::ConnectionType conn
     stream.state = Stream::closed;
     auto httpReply = stream.reply();
     if (httpReply) {
+        int statusCode = httpReply->statusCode();
+        if (statusCode == 401 || statusCode == 407) {
+            // handleAuthorization will either re-send the request or
+            // finishWithError. In either case we don't want to emit finished
+            // here.
+            handleAuthorization(stream);
+            return;
+        }
+
         httpReply->disconnect(this);
         if (stream.data())
             stream.data()->disconnect(this);
@@ -1473,54 +1487,12 @@ quint32 QHttp2ProtocolHandler::allocateStreamID()
     return streamID;
 }
 
-static std::optional<QUrl> makeUrl(const HPack::HttpHeader &requestHeader)
-{
-    constexpr QByteArrayView names[] = { ":authority", ":method", ":path", ":scheme" };
-    enum PseudoHeaderEnum
-    {
-        Authority,
-        Method,
-        Path,
-        Scheme
-    };
-    std::array<std::optional<QByteArrayView>, std::size(names)> pseudoHeaders{};
-    for (const auto &field : requestHeader) {
-        const auto it = std::find(std::begin(names), std::end(names), QByteArrayView(field.name));
-        if (it != std::end(names)) {
-            const auto index = std::distance(std::begin(names), it);
-            if (field.value.isEmpty() || pseudoHeaders.at(index).has_value())
-                return {};
-            pseudoHeaders[index] = field.value;
-        }
-    }
-
-    if (!std::all_of(pseudoHeaders.begin(), pseudoHeaders.end(), [](const auto &x) { return x.has_value();})) {
-        // All four required, HTTP/2 8.1.2.3.
-        return {};
-    }
-
-    const QByteArrayView method = pseudoHeaders[Method].value();
-    if (method.compare("get", Qt::CaseInsensitive) != 0 &&
-        method.compare("head", Qt::CaseInsensitive) != 0) {
-        return {};
-    }
-
-    QUrl url;
-    url.setScheme(QLatin1StringView(pseudoHeaders[Scheme].value()));
-    url.setAuthority(QLatin1StringView(pseudoHeaders[Authority].value()));
-    url.setPath(QLatin1StringView(pseudoHeaders[Path].value()));
-
-    if (!url.isValid())
-        return {};
-    return url;
-}
-
 bool QHttp2ProtocolHandler::tryReserveStream(const Http2::Frame &pushPromiseFrame,
                                              const HPack::HttpHeader &requestHeader)
 {
     Q_ASSERT(pushPromiseFrame.type() == FrameType::PUSH_PROMISE);
 
-    const auto url = makeUrl(requestHeader);
+    const auto url = HPack::makePromiseKeyUrl(requestHeader);
     if (!url.has_value())
         return false;
 
