@@ -62,6 +62,8 @@ private slots:
 
     void hideNativeByDestruction();
 
+    void explicitDoneAfterButtonClicked();
+
     void cleanup();
 };
 
@@ -437,7 +439,7 @@ void tst_QMessageBox::shortcut()
     msgBox.addButton("&Maybe", QMessageBox::YesRole);
     ExecCloseHelper closeHelper;
     closeHelper.start(Qt::Key_M, &msgBox);
-    QCOMPARE(msgBox.exec(), 2);
+    QCOMPARE(msgBox.exec(), 4);
 }
 #endif
 
@@ -509,7 +511,7 @@ QT_WARNING_DISABLE_DEPRECATED
     // the button text versions
     closeHelper.start(Qt::Key_Enter);
     ret = QMessageBox::information(nullptr, "title", "text", "Yes", "No", QString(), 1);
-    COMPARE(ret, 1);
+    COMPARE(ret, 3); // Custom button opaque result
     QVERIFY(closeHelper.done());
 #endif // QT_DEPRECATED_SINCE(6, 2)
 #undef COMPARE
@@ -538,9 +540,9 @@ void tst_QMessageBox::instanceSourceCompat()
 #ifndef Q_OS_MAC
     // mnemonics are not used on OS X
     closeHelper.start(QKeyCombination(Qt::ALT | Qt::Key_R).toCombined(), &mb);
-    QCOMPARE(mb.exec(), 0);
+    QCOMPARE(mb.exec(), 2);
     closeHelper.start(QKeyCombination(Qt::ALT | Qt::Key_Z).toCombined(), &mb);
-    QCOMPARE(mb.exec(), 1);
+    QCOMPARE(mb.exec(), 3);
 #endif
 }
 
@@ -815,5 +817,51 @@ void tst_QMessageBox::hideNativeByDestruction()
     QVERIFY(QTest::qWaitFor(windowActive));
 }
 
+void tst_QMessageBox::explicitDoneAfterButtonClicked()
+{
+    QMessageBox msgBox;
+    auto *standardButton = msgBox.addButton(QMessageBox::Ok);
+    auto *customButton = msgBox.addButton("Custom", QMessageBox::RejectRole);
+
+    QSignalSpy acceptedSpy(&msgBox, &QDialog::accepted);
+    QSignalSpy rejectedSpy(&msgBox, &QDialog::rejected);
+
+    msgBox.setDefaultButton(standardButton);
+    ExecCloseHelper closeHelper;
+    closeHelper.start(Qt::Key_Enter, &msgBox);
+    msgBox.exec();
+    QCOMPARE(msgBox.clickedButton(), standardButton);
+    QCOMPARE(msgBox.result(), QMessageBox::Ok);
+    QCOMPARE(acceptedSpy.size(), 1);
+    QCOMPARE(rejectedSpy.size(), 0);
+
+    msgBox.accept();
+    QCOMPARE(msgBox.result(), QDialog::Accepted);
+    QCOMPARE(acceptedSpy.size(), 2);
+    QCOMPARE(rejectedSpy.size(), 0);
+    msgBox.reject();
+    QCOMPARE(msgBox.result(), QDialog::Rejected);
+    QCOMPARE(acceptedSpy.size(), 2);
+    QCOMPARE(rejectedSpy.size(), 1);
+
+    msgBox.setDefaultButton(customButton);
+    closeHelper.start(Qt::Key_Enter, &msgBox);
+    msgBox.exec();
+    QCOMPARE(msgBox.clickedButton(), customButton);
+    QVERIFY(msgBox.result() != QDialog::Accepted);
+    QVERIFY(msgBox.result() != QDialog::Rejected);
+    QCOMPARE(acceptedSpy.size(), 2);
+    QCOMPARE(rejectedSpy.size(), 2);
+
+    msgBox.accept();
+    QCOMPARE(msgBox.result(), QDialog::Accepted);
+    QCOMPARE(acceptedSpy.size(), 3);
+    QCOMPARE(rejectedSpy.size(), 2);
+    msgBox.reject();
+    QCOMPARE(msgBox.result(), QDialog::Rejected);
+    QCOMPARE(acceptedSpy.size(), 3);
+    QCOMPARE(rejectedSpy.size(), 3);
+}
+
 QTEST_MAIN(tst_QMessageBox)
 #include "tst_qmessagebox.moc"
